import { Animation } from 'core/Animation';
import Coordinate from 'geo/Coordinate';
import Point from 'geo/Point';
import Map from './Map';
import { isNil, isFunction, hasOwn } from 'core/util';

function equalView(view1, view2) {
    for (const p in view1) {
        if (hasOwn(view1, p)) {
            if (p === 'center') {
                if (view1[p][0] !== view2[p][0] || view1[p][1] !== view2[p][1]) {
                    return false;
                }
            } else if (view1[p] !== view2[p]) {
                return false;
            }
        }
    }
    return true;
}

Map.include(/** @lends Map.prototype */{

    /**
     * Update map's view with animation.
     * @example
     * map.animateTo({
     *     zoom : 13,
     *     center : [0, 0],
     *     pitch : 30,
     *     bearing : 60
     * }, {
     *     duration : 6000,
     *     easing : 'out'
     * }, function(frame) {
     *     if (frame.state.playState === 'finished') {
     *         console.log('animation finished');
     *     }
     * });
     * @param  {Object} view    view object
     * @param  {Object} [options=null]
     * @param  {String} [options.easing=out]
     * @param  {Number} [options.duration=map.options.zoomAnimationDuration]
     * @param  {Function} [step=null]  - step function during animation, animation frame as the parameter
     * @return {Map}         this
     */
    animateTo(view, options = {}, step) {
        this._stopAnim();
        if (isFunction(options)) {
            step = options;
            options = {};
        }
        const projection = this.getProjection(),
            currView = this.getView(),
            props = {};
        let empty = true;
        for (const p in view) {
            if (hasOwn(view, p) && !isNil(currView[p])) {
                empty = false;
                if (p === 'center') {
                    const from = new Coordinate(currView[p]).toFixed(7),
                        to = new Coordinate(view[p]).toFixed(7);
                    if (!from.equals(to)) {
                        props['center'] = [from, to];
                    }
                } else if (currView[p] !== view[p] && p !== 'around') {
                    props[p] = [currView[p], view[p]];
                }
            }
        }
        if (empty) {
            return this;
        }
        const zoomOrigin = view['around'] || new Point(this.width / 2, this.height / 2);
        let preView = this.getView();
        const renderer = this._getRenderer(),
            framer = function (fn) {
                renderer.callInNextFrame(fn);
            };

        const player = this._animPlayer = Animation.animate(props, {
            'easing': options['easing'] || 'out',
            'duration': options['duration'] || this.options['zoomAnimationDuration'],
            'framer' : framer
        }, frame => {
            if (this.isRemoved()) {
                player.finish();
                return;
            }
            if (player.playState === 'running') {
                const view = this.getView();
                if (!equalView(view, preView)) {
                    // map's view is updated and stop animation
                    this._stopAnim();
                    return;
                }
                if (frame.styles['center']) {
                    const center = frame.styles['center'];
                    this._setPrjCenter(projection.project(center));
                    this.onMoving(this._parseEventFromCoord(this.getCenter()));
                }
                if (!isNil(frame.styles['zoom'])) {
                    this.onZooming(frame.styles['zoom'], zoomOrigin);
                }
                if (!isNil(frame.styles['pitch'])) {
                    this.setPitch(frame.styles['pitch']);
                }
                if (!isNil(frame.styles['bearing'])) {
                    this.setBearing(frame.styles['bearing']);
                }
                preView = this.getView();
                this._fireEvent('animating');
            } else if (player.playState === 'finished') {
                if (!player._interupted) {
                    if (props['center']) {
                        this._setPrjCenter(projection.project(props['center'][1]));
                    }
                    if (!isNil(props['pitch'])) {
                        this.setPitch(props['pitch'][1]);
                    }
                    if (!isNil(props['bearing'])) {
                        this.setBearing(props['bearing'][1]);
                    }
                }
                this._endAnim(props, zoomOrigin, options);
            }
            if (step) {
                step(frame);
            }
        });

        this._startAnim(props, zoomOrigin);

        return this;
    },

    /**
     * Whether the map is animating with .animateTo
     * @return {Boolean}
     */
    isAnimating() {
        return !!(this._animPlayer);
    },

    isRotating() {
        return this.isDragRotating() || !!this._animRotating;
    },

    _endAnim(props, zoomOrigin, options) {
        delete this._animRotating;
        let evtType;
        if (this._animPlayer) {
            evtType = this._animPlayer._interupted ? 'animateinterupted' : 'animateend';
            delete this._animPlayer;
        }
        if (props['center']) {
            this.onMoveEnd(this._parseEventFromCoord(props['center'][1]));
        }
        if (!isNil(props['zoom'])) {
            if (!options['wheelZoom']) {
                this.onZoomEnd(props['zoom'][1], zoomOrigin);
            } else {
                this.onZooming(props['zoom'][1], zoomOrigin);
            }
        }
        if (evtType) {
            this._fireEvent(evtType);
        }
    },

    _startAnim(props, zoomOrigin) {
        if (!this._animPlayer) {
            return;
        }
        if (props['center']) {
            this.onMoveStart();
        }
        if (props['zoom'] && !this.isZooming()) {
            this.onZoomStart(props['zoom'][1], zoomOrigin);
        }
        if (props['pitch'] || props['bearing']) {
            this._animRotating = true;
        }
        this._fireEvent('animatestart');
        this._animPlayer.play();
    },

    _stopAnim() {
        if (this._animPlayer) {
            this._animPlayer._interupted = true;
            this._animPlayer.finish();
        }
    }
});
